#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>

#include "../include/os.h"

#include "gw.h"
#include "menu.h"
#include "sound.h"
#include "sar.h"


static sar_menu_spin_struct *SAROptionsGetSpinByID(
	sar_menu_struct *m, int id_code, int *n
);
static sar_menu_switch_struct *SAROptionsGetSwitchByID(
	sar_menu_struct *m, int id_code, int *n
);

static void SARMenuOptionsReinitJoysticks(sar_core_struct *core_ptr);
static void SARMenuOptionsRemapJoystickButtons(sar_core_struct *core_ptr);
#ifdef JS_SUPPORT
static int SARGetJSButtonFromSpinAction(
        sar_menu_spin_struct *act_spin_ptr, int id_code
);
#endif	/* JS_SUPPORT */

void SARApplyOptionsFromMenus(sar_core_struct *core_ptr);
void SARFetchOptionsToMenus(sar_core_struct *core_ptr);

void SARMenuOptionsSwitchCB(
	void *object, void *client_data,
	int id_code, Boolean state
);
void SARMenuOptionsSpinCB(
	void *object, void *client_data,
        int id_code, char *value
);


/*
 *	Searches for the first spin object on menu m with a matching id
 *	code, returns the pointer to the spin object or NULL on no match.
 */
static sar_menu_spin_struct *SAROptionsGetSpinByID(
	sar_menu_struct *m, int id_code, int *n
)
{
	int i;
	void *ptr;
	sar_menu_spin_struct *spin;


        if(n != NULL) 
            (*n) = -1;

	if(m == NULL)
	    return(NULL);

	for(i = 0; i < m->total_objects; i++)
	{
	    ptr = m->object[i];
	    if(ptr == NULL)
		continue;

	    if(*(int *)ptr == SAR_MENU_OBJECT_TYPE_SPIN)
		spin = ptr;
	    else
		continue;

	    if(spin->id_code == id_code)
	    {
		if(n != NULL)
		    (*n) = i;

		return(spin);
	    }
	}

	return(NULL);
}

/*
 *	Searches for the first switch object on menu m with a matching id
 *      code, returns the pointer to the switch object or NULL on no match.
 */
static sar_menu_switch_struct *SAROptionsGetSwitchByID(
        sar_menu_struct *m, int id_code, int *n
)
{
        int i;
        void *ptr;
        sar_menu_switch_struct *switch_ptr;


	if(n != NULL)
	    (*n) = -1;

        if(m == NULL)
            return(NULL);

        for(i = 0; i < m->total_objects; i++)
        {
            ptr = m->object[i];
            if(ptr == NULL)
                continue;
 
            if(*(int *)ptr == SAR_MENU_OBJECT_TYPE_SWITCH)
                switch_ptr = ptr;
            else
                continue;

            if(switch_ptr->id_code == id_code)
	    {
                if(n != NULL)
                    (*n) = i;

                return(switch_ptr);
	    }
        }

        return(NULL);
}


/*
 *	Reinitializes the joystick and the game controller based on
 *	the global options.
 *
 *	This will update the gctl structure on the core structure and
 *	option.gctl_controllers.
 */
static void SARMenuOptionsReinitJoysticks(sar_core_struct *core_ptr)
{
	if(core_ptr == NULL)
	    return;

#ifdef JS_SUPPORT
	/* Set up game controllers mask.
	 *
	 * Start off game controllers mask with just keyboard and pointer.
	 */
	option.gctl_controllers = GCTL_CONTROLLER_KEYBOARD |
	    GCTL_CONTROLLER_POINTER;
	/* Check if one or more joystick(s) need to be enabled. */
	if(option.gctl_js0_axis_roles ||
	   option.gctl_js1_axis_roles
	)
	{
	    /* Add joystick to game controllers mask. */
	    option.gctl_controllers |= GCTL_CONTROLLER_JOYSTICK;
	}

	/* Reinitialize game controller to the new type. */
	GWSetInputBusy(core_ptr->display);
	SARInitGCTL(core_ptr);		/* Realize changes. */
	GWSetInputReady(core_ptr->display);
#else   /* JS_SUPPORT */
	/* Print warning about no joystick support. */
	GWOutputMessage(
	    core_ptr->display, GWOutputMessageTypeError,
	    "No Joystick Support Available",
"There is no joystick support available.",
"This program was not compiled with joystick support and/or\n\
joystick support was not enabled on your operating system."
	);

	/* Reinitialize game controler back to keyboard and pointer since
	 * joystick support is not compiled.
	 */
	option.gctl_controllers = GCTL_CONTROLLER_KEYBOARD |
	    GCTL_CONTROLLER_POINTER;
	SARInitGCTL(core_ptr);

#endif  /* !JS_SUPPORT */
}

/*
 *	Checks if the gctl structure on the given core structure has
 *	one or more joysticks initialized. If so then the joystick
 *	structures on the gctl structure will have its button mappings
 *	updated with the global options joystick button mappings.
 *
 *	No reinitialization will take place.
 */
static void SARMenuOptionsRemapJoystickButtons(sar_core_struct *core_ptr)
{
	gctl_struct *gc;
	sar_option_struct *opt;

        if(core_ptr == NULL)
            return;

	gc = core_ptr->gctl;
	if(gc == NULL)
	    return;

	opt = &option;

#ifdef JS_SUPPORT
	/* Joystick controller defined as one of the game controllers? */
	if(gc->controllers & GCTL_CONTROLLER_JOYSTICK)
	{
	    int i;
	    gctl_js_struct *gc_js;


	    /* First joystick initialized? */
	    i = 0;
	    if(i < gc->total_joysticks)
		gc_js = &gc->joystick[i];
	    else
		gc_js = NULL;
	    if(gc_js != NULL)
	    {
		gc_js->button_rotate = opt->js0_btn_rotate;
		gc_js->button_air_brakes = opt->js0_btn_air_brakes;
                gc_js->button_wheel_brakes =  opt->js0_btn_wheel_brakes;
                gc_js->button_zoom_in = opt->js0_btn_zoom_in;
                gc_js->button_zoom_out = opt->js0_btn_zoom_out;
                gc_js->button_hoist_up = opt->js0_btn_hoist_up;
                gc_js->button_hoist_down = opt->js0_btn_hoist_down;
	    }

            /* Second joystick initialized? */
            i = 1;
            if(i < gc->total_joysticks)
                gc_js = &gc->joystick[i];
            else
                gc_js = NULL;
            if(gc_js != NULL)
            {
                gc_js->button_rotate = opt->js1_btn_rotate;
                gc_js->button_air_brakes = opt->js1_btn_air_brakes;
                gc_js->button_wheel_brakes =  opt->js1_btn_wheel_brakes;
                gc_js->button_zoom_in = opt->js1_btn_zoom_in;
                gc_js->button_zoom_out = opt->js1_btn_zoom_out;
                gc_js->button_hoist_up = opt->js1_btn_hoist_up;
                gc_js->button_hoist_down = opt->js1_btn_hoist_down;
            }

	}
#endif  /* !JS_SUPPORT */
}


#ifdef JS_SUPPORT
/*
 *	Returns the current joystick button mapping for the given
 *	joystick action spin's current value.
 *
 *	The given id_code must be one of SAR_OPT_SELECT_JS#_BUTTON_ACTION
 *	where # is the joystick number.
 *
 *	Can return -1 on error or if the value is not set. Calling
 *	function must add 1 to offset when setting to the button spin's
 *	current value since the button spin's value 0 means button number
 *	-1 (none/invalid).
 */
static int SARGetJSButtonFromSpinAction(
	sar_menu_spin_struct *act_spin_ptr, int id_code
)
{
	int val = -1;
	sar_option_struct *opt = &option;


	if(act_spin_ptr == NULL)
	    return(val);

	/* First joystick (js0). */
	if(id_code == SAR_OPT_SELECT_JS0_BUTTON_ACTION)
	{
	    switch(act_spin_ptr->cur_value)
	    {
	      case 0:	/* Rotate Modifier. */
		val = opt->js0_btn_rotate;
		break;
	      case 1:	/* Air Brakes. */
		val = opt->js0_btn_air_brakes;
		break;
	      case 2:	/* Wheel Brakes. */
		val = opt->js0_btn_wheel_brakes;
		break;
	      case 3:	/* Zoom In. */
		val = opt->js0_btn_zoom_in;
		break;
	      case 4:	/* Zoom out. */
		val = opt->js0_btn_zoom_out;
		break;
	      case 5:	/* Hoist Up. */
		val = opt->js0_btn_hoist_up;
		break;
	      case 6:	/* Hoist Down. */
		val = opt->js0_btn_hoist_down;
		break;
/* When adding a new button mapping, besure to add it to the setting of
   the button mappings to option structure in SARMenuOptionsSpinCB().

   Also be sure to add support for EACH joystick!
 */
	    }
	}
	/* Second joystick (js1). */
        else if(id_code == SAR_OPT_SELECT_JS1_BUTTON_ACTION)
        {
            switch(act_spin_ptr->cur_value) 
            {
              case 0:   /* Rotate Modifier. */
                val = opt->js1_btn_rotate;
                break;
              case 1:   /* Air Brakes. */
                val = opt->js1_btn_air_brakes;
                break;
              case 2:   /* Wheel Brakes. */
                val = opt->js1_btn_wheel_brakes;
                break;
              case 3:   /* Zoom In. */
                val = opt->js1_btn_zoom_in;
                break;
              case 4:   /* Zoom out. */
                val = opt->js1_btn_zoom_out;
                break;
              case 5:   /* Hoist Up. */
                val = opt->js1_btn_hoist_up;
                break;
              case 6:   /* Hoist Down. */
                val = opt->js1_btn_hoist_down;
                break;
/* When adding a new button mapping, besure to add it to the setting of   
   the button mappings to option structure in SARMenuOptionsSpinCB().

    Also be sure to add support for EACH joystick!
 */
            }
	}

	if(val < 0)
	    val = -1;

	return(val);
}
#endif	/* JS_SUPPORT */

/*
 *	Sets global option structure values based on values from the
 *	options menu widget/objects.
 */
void SARApplyOptionsFromMenus(sar_core_struct *core_ptr)
{
	int i;
	sar_menu_struct *m;


	if(core_ptr == NULL)
	    return;

	/* Options general. */
	i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_OPTIONS);
	m = ((i < 0) ? NULL : core_ptr->menu[i]);
	if(m != NULL)
	{

	}

        /* Options controller. */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_OPTIONS_CONTROLLER);
        m = ((i < 0) ? NULL : core_ptr->menu[i]);
        if(m != NULL)
        {

        }

        /* Options graphics. */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_OPTIONS_GRAPHICS);
        m = ((i < 0) ? NULL : core_ptr->menu[i]);
        if(m != NULL)
        {

        }

        /* Options sounds. */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_OPTIONS_SOUND);
        m = ((i < 0) ? NULL : core_ptr->menu[i]);
        if(m != NULL)
        {

        }
}

/*
 *	Gets values from global options and sets them to the objects in
 *	the widget/objects on the options menus.
 */
void SARFetchOptionsToMenus(sar_core_struct *core_ptr)
{
        int i;
        sar_menu_struct *m;
	sar_menu_spin_struct *spin_ptr;
	sar_menu_switch_struct *switch_ptr;


        if(core_ptr == NULL)
            return;

        /* Options: General */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_OPTIONS);
        m = ((i < 0) ? NULL : core_ptr->menu[i]);
        if(m != NULL)
        {

	}

        /* Options: Simulation (contact sizes, flight dynamics, damage) */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_OPTIONS_SIMULATION);
        m = ((i < 0) ? NULL : core_ptr->menu[i]);
        if(m != NULL)
        {
            /* Hoist contact, the expansion coefficient from hoist rescue
	     * basket's actual size.
	     */
            spin_ptr = SAROptionsGetSpinByID(
		m, SAR_OPT_SELECT_HOIST_CONTACT, NULL
	    );
            if(spin_ptr != NULL)
            {
                if(option.hoist_contact_expansion_coeff >= 4.0)
		    spin_ptr->cur_value = 0;
                else if(option.hoist_contact_expansion_coeff >= 2.0)
                    spin_ptr->cur_value = 1;
                else if(option.hoist_contact_expansion_coeff >= 1.0)
                    spin_ptr->cur_value = 2;
		else
		    spin_ptr->cur_value = 0;
	    }

            /* Damage resistance, how much of an impact can player
	     * aircraft tolorate from actual coefficient.
	     */
            spin_ptr = SAROptionsGetSpinByID(
		m, SAR_OPT_SELECT_DAMAGE_RESISTANCE, NULL
	    );
            if(spin_ptr != NULL)
            {
		if(option.damage_resistance_coeff >= 2.0)
		    spin_ptr->cur_value = 0;
		else if(option.damage_resistance_coeff >= 1.5)
		    spin_ptr->cur_value = 1;
                else
                    spin_ptr->cur_value = 2;
            }

            /* Flight dynamics dificulty (easy to realistic). */
            spin_ptr = SAROptionsGetSpinByID(
		m, SAR_OPT_SELECT_FLIGHT_PHYSICS, NULL
	    );
            if(spin_ptr != NULL)
            {
		switch(option.flight_physics)
		{
		  case FLIGHT_PHYSICS_REALISTIC:
		    spin_ptr->cur_value = 2;
		    break;

		  case FLIGHT_PHYSICS_MODERATE:
		    spin_ptr->cur_value = 1;
                    break;

		  default:
		    spin_ptr->cur_value = 0;
		    break;
		}
            }
	}


#ifdef JS_SUPPORT
        /* Options: Controller (joysticks) */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_OPTIONS_CONTROLLER);
        m = ((i < 0) ? NULL : core_ptr->menu[i]);
        if(m != NULL)
        {
	    int connection;
	    gctl_flags_t axis_roles;


	    /* First joystick (js0) connection type and axis roles. */
	    /* Connection type. */
	    connection = option.js0_connection;
            spin_ptr = SAROptionsGetSpinByID(
                m, SAR_OPT_SELECT_JS0_CONNECTION, NULL
            );
            if(spin_ptr != NULL)
            {
		switch(connection)
		{
		  case GCTL_JS_CONNECTION_USB:
		    spin_ptr->cur_value = 1;
		    break;
		  default:	/* GCTL_JS_CONNECTION_STANDARD */
		    spin_ptr->cur_value = 0;
                    break;
		}
	    }
	    /* Axis roles. */
	    axis_roles = option.gctl_js0_axis_roles;
            spin_ptr = SAROptionsGetSpinByID(
		m, SAR_OPT_SELECT_JS0_AXISES, NULL
	    );
            if(spin_ptr != NULL)
            {
		/* These value codes should match those in SARBuildMenus()
		 * that added these values.
		 */
                if(axis_roles)
                {
		    if(axis_roles & GCTL_JS_AXIS_ROLE_AS_THROTTLE_AND_RUDDER)
			spin_ptr->cur_value = 9;
                    else if((axis_roles & GCTL_JS_AXIS_ROLE_THROTTLE) &&
                            (axis_roles & GCTL_JS_AXIS_ROLE_HEADING) &&
                            (axis_roles & GCTL_JS_AXIS_ROLE_HAT)
                    )
                        spin_ptr->cur_value = 8;
                    else if((axis_roles & GCTL_JS_AXIS_ROLE_HEADING) &&
                            (axis_roles & GCTL_JS_AXIS_ROLE_HAT)
                    )
                        spin_ptr->cur_value = 7;
                    else if((axis_roles & GCTL_JS_AXIS_ROLE_HEADING) &&
                            (axis_roles & GCTL_JS_AXIS_ROLE_THROTTLE)
                    )
                        spin_ptr->cur_value = 6;
                    else if(axis_roles & GCTL_JS_AXIS_ROLE_HEADING)
                        spin_ptr->cur_value = 5;
                    else if((axis_roles & GCTL_JS_AXIS_ROLE_THROTTLE) &&
                            (axis_roles & GCTL_JS_AXIS_ROLE_HAT)
                    )
                        spin_ptr->cur_value = 4;
                    else if(axis_roles & GCTL_JS_AXIS_ROLE_HAT)
                        spin_ptr->cur_value = 3;
                    else if(axis_roles & GCTL_JS_AXIS_ROLE_THROTTLE) 
                        spin_ptr->cur_value = 2;
                    else
                        spin_ptr->cur_value = 1;
                }
                else
                {
                    spin_ptr->cur_value = 0;
                }
            }

	    /* Second joystick (js1) connection type and axis roles. */
            /* Connection type. */
            connection = option.js1_connection;
            spin_ptr = SAROptionsGetSpinByID(
                m, SAR_OPT_SELECT_JS1_CONNECTION, NULL
            );
            if(spin_ptr != NULL)
            {
                switch(connection)
                {
                  case GCTL_JS_CONNECTION_USB:
                    spin_ptr->cur_value = 1;
                    break;
                  default:      /* GCTL_JS_CONNECTION_STANDARD */
                    spin_ptr->cur_value = 0;
                    break;
                }
            }
	    /* Axis roles. */
            axis_roles = option.gctl_js1_axis_roles;
            spin_ptr = SAROptionsGetSpinByID(
		m, SAR_OPT_SELECT_JS1_AXISES, NULL
	    );
            if(spin_ptr != NULL)
            {
                /* These value codes should match those in SARBuildMenus()
                 * that added these values.
                 */
                if(axis_roles)
                {
                    if(axis_roles & GCTL_JS_AXIS_ROLE_AS_THROTTLE_AND_RUDDER)
                        spin_ptr->cur_value = 9;
                    else if((axis_roles & GCTL_JS_AXIS_ROLE_THROTTLE) &&
                            (axis_roles & GCTL_JS_AXIS_ROLE_HEADING) &&
                            (axis_roles & GCTL_JS_AXIS_ROLE_HAT)
                    )
                        spin_ptr->cur_value = 8;
                    else if((axis_roles & GCTL_JS_AXIS_ROLE_HEADING) &&
                            (axis_roles & GCTL_JS_AXIS_ROLE_HAT)
                    )
                        spin_ptr->cur_value = 7;
                    else if((axis_roles & GCTL_JS_AXIS_ROLE_HEADING) &&
                            (axis_roles & GCTL_JS_AXIS_ROLE_THROTTLE)
                    )
                        spin_ptr->cur_value = 6;
                    else if(axis_roles & GCTL_JS_AXIS_ROLE_HEADING)
                        spin_ptr->cur_value = 5;
                    else if((axis_roles & GCTL_JS_AXIS_ROLE_THROTTLE) &&
                            (axis_roles & GCTL_JS_AXIS_ROLE_HAT)
                    )
                        spin_ptr->cur_value = 4;
                    else if(axis_roles & GCTL_JS_AXIS_ROLE_HAT)
                        spin_ptr->cur_value = 3;
                    else if(axis_roles & GCTL_JS_AXIS_ROLE_THROTTLE)
                        spin_ptr->cur_value = 2;
                    else
                        spin_ptr->cur_value = 1;
                }
		else
		{
		    spin_ptr->cur_value = 0;
		}
            }
        }
#endif	/* JS_SUPPORT */


#ifdef JS_SUPPORT
        /* Options: Controller joystick button mappings. */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_OPTIONS_CONTROLLER_JS_BTN);
        m = ((i < 0) ? NULL : core_ptr->menu[i]);
        if(m != NULL)
        {
	    sar_menu_spin_struct *act_spin_ptr;


            /* First joystick (js0) button actions. */
            act_spin_ptr = SAROptionsGetSpinByID(
		m, SAR_OPT_SELECT_JS0_BUTTON_ACTION, NULL
	    );
	    spin_ptr = SAROptionsGetSpinByID(
                m, SAR_OPT_SELECT_JS0_BUTTON_NUMBER, NULL
            );
	    if((spin_ptr != NULL) && (act_spin_ptr != NULL))
	    {
		spin_ptr->cur_value = SARGetJSButtonFromSpinAction(
		    act_spin_ptr, SAR_OPT_SELECT_JS0_BUTTON_ACTION
		) + 1;
		if(spin_ptr->cur_value >= spin_ptr->total_values)
		    spin_ptr->cur_value = 0;
	    }

            /* Second joystick (js1) button actions. */
            act_spin_ptr = SAROptionsGetSpinByID(  
                m, SAR_OPT_SELECT_JS1_BUTTON_ACTION, NULL
            );
            spin_ptr = SAROptionsGetSpinByID(
                m, SAR_OPT_SELECT_JS1_BUTTON_NUMBER, NULL
            );
            if((spin_ptr != NULL) && (act_spin_ptr != NULL))
            {
		spin_ptr->cur_value = SARGetJSButtonFromSpinAction(
                    act_spin_ptr, SAR_OPT_SELECT_JS1_BUTTON_ACTION
                ) + 1;
                if(spin_ptr->cur_value >= spin_ptr->total_values)
                    spin_ptr->cur_value = 0;
	    }
	}
#endif	/* JS_SUPPORT */

        /* Options graphics. */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_OPTIONS_GRAPHICS);
        m = ((i < 0) ? NULL : core_ptr->menu[i]);
        if(m != NULL)
        {
	    /* Textured ground switch. */
            switch_ptr = SAROptionsGetSwitchByID(
		m, SAR_OPT_SELECT_GROUND_TEXTURE, NULL
	    );
            if(switch_ptr != NULL)
                switch_ptr->state = option.textured_ground;

	    /* Textured objects switch. */
            switch_ptr = SAROptionsGetSwitchByID(
		m, SAR_OPT_SELECT_OBJECT_TEXTURE, NULL
	    );
            if(switch_ptr != NULL)
                switch_ptr->state = option.textured_objects;

	    /* Textured clouds switch. */
            switch_ptr = SAROptionsGetSwitchByID(
		m, SAR_OPT_SELECT_CLOUDS, NULL
	    );
            if(switch_ptr != NULL)
                switch_ptr->state = option.textured_clouds;

	    /* Atmosphere switch. */
	    switch_ptr = SAROptionsGetSwitchByID(
		m, SAR_OPT_SELECT_ATMOSPHERE, NULL
	    );
            if(switch_ptr != NULL)
                switch_ptr->state = option.atmosphere;

	    /* Dual pass depth switch. */
            switch_ptr = SAROptionsGetSwitchByID(
                m, SAR_OPT_SELECT_DUAL_PASS_DEPTH, NULL
            );
            if(switch_ptr != NULL)
                switch_ptr->state = option.dual_pass_depth;

	    /* Prop wash switch. */
	    switch_ptr = SAROptionsGetSwitchByID(
                m, SAR_OPT_SELECT_PROP_WASH, NULL
            );
            if(switch_ptr != NULL)
                switch_ptr->state = option.prop_wash;

            /* Smoke trails switch. */
            switch_ptr = SAROptionsGetSwitchByID(
                m, SAR_OPT_SELECT_SMOKE_TRAILS, NULL
            );
            if(switch_ptr != NULL)
                switch_ptr->state = option.smoke_trails;

            /* Celestial objects switch. */
            switch_ptr = SAROptionsGetSwitchByID(
                m, SAR_OPT_SELECT_CELESTIAL_OBJECTS, NULL
            );
            if(switch_ptr != NULL)
                switch_ptr->state = option.celestial_objects;


            spin_ptr = SAROptionsGetSpinByID(
		m, SAR_OPT_SELECT_MAX_VISIBILITY, NULL
	    );
            if(spin_ptr != NULL)
            {
		int i = option.visibility_max;

		if((i >= 0) && (i < spin_ptr->total_values))
		    spin_ptr->cur_value = i;
		else
		    spin_ptr->cur_value = 0;
            } 
        }

        /* Options sounds. */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_OPTIONS_SOUND);
        m = ((i < 0) ? NULL : core_ptr->menu[i]);
        if(m != NULL)
        {
	    /* Sounds level. */
	    spin_ptr = SAROptionsGetSpinByID(
		m, SAR_OPT_SELECT_SOUND_LEVEL, NULL
	    );
	    if(spin_ptr != NULL)
	    {
		/* No recorder (implies all sounds off)? */
		if(core_ptr->recorder == NULL)
		{
		    spin_ptr->cur_value = 0;
		}
		/* All sounds on? */
		else if(option.engine_sounds &&
                        option.event_sounds &&
                        option.voice_sounds
		)
		{
		    spin_ptr->cur_value = 3;
		}
		/* Engine sounds only? */
		else if(option.engine_sounds)
		{
		    spin_ptr->cur_value = 2;
		}
		/* Event sounds only? */
		else if(option.event_sounds)
		{
		    spin_ptr->cur_value = 1;
		}
		/* All else no sounds. */
		else
		{
		    spin_ptr->cur_value = 0;
		}
	    }

	    /* Music. */
            switch_ptr = SAROptionsGetSwitchByID(
		m, SAR_OPT_SELECT_MUSIC, NULL
	    );
	    if(switch_ptr != NULL)
		switch_ptr->state = option.music;
        }
}

/*         
 *      Options menu select switch callback.
 */
void SARMenuOptionsSwitchCB(
        void *object, void *client_data,
        int id_code, Boolean state
)
{
        sar_core_struct *core_ptr = (sar_core_struct *)client_data;
        sar_menu_switch_struct *switch_ptr = (sar_menu_switch_struct *)object;


        if((core_ptr == NULL) || (switch_ptr == NULL))
            return;

        /* Handle by switch's ID code. */
        switch(id_code)
        {
          /* ****************************************************** */
          case SAR_OPT_SELECT_GROUND_TEXTURE:
            option.textured_ground = state;
            break;

          /* ****************************************************** */
          case SAR_OPT_SELECT_OBJECT_TEXTURE:
            option.textured_objects = state;
	    if(!state)
	    {
		/* Print warning about object texture if off. */
		GWOutputMessage(
		    core_ptr->display, GWOutputMessageTypeWarning,
		    "Object Texture Disabling Warning",
"Disabling object texturing may affect ground plot\n\
objects and make flight visualization extremely difficult.",
"Ground plot objects and visualization important objects\n\
will not be texturized. It will also be extremely difficult\n\
to visually interprite heightfield objects."
		);
	    }
            break;

          /* ****************************************************** */
          case SAR_OPT_SELECT_CLOUDS:
            option.textured_clouds = state;
            break;

          /* ****************************************************** */
          case SAR_OPT_SELECT_ATMOSPHERE:
            option.atmosphere = state;  
            break;
 
          /* ****************************************************** */
          case SAR_OPT_SELECT_DUAL_PASS_DEPTH:
            option.dual_pass_depth = state;
            break;

          /* ****************************************************** */
          case SAR_OPT_SELECT_PROP_WASH:
            option.prop_wash = state;
            break;

          /* ****************************************************** */
          case SAR_OPT_SELECT_SMOKE_TRAILS:
            option.smoke_trails = state;
            break;

          /* ****************************************************** */
          case SAR_OPT_SELECT_CELESTIAL_OBJECTS:
            option.celestial_objects = state;
            break;

          /* ****************************************************** */
          case SAR_OPT_SELECT_MUSIC:
            option.music = state;
            break;
        }
}
 
/*
 *      Options menu select spin callback.
 */
void SARMenuOptionsSpinCB(
        void *object, void *client_data,
        int id_code, char *value
)
{               
        sar_core_struct *core_ptr = (sar_core_struct *)client_data;
        sar_menu_spin_struct *spin_ptr = (sar_menu_spin_struct *)object;
        sar_menu_struct *m;
        const char *sound_server_connect_arg = "127.0.0.1:9433";


        if((core_ptr == NULL) ||
           (spin_ptr == NULL)
        )
            return;

	/* Recorder address available? */
	if(core_ptr->recorder_address != NULL)
	    sound_server_connect_arg = (const char *)core_ptr->recorder_address;

	/* Get current menu pointer (must be valid). */
	m = SARGetCurrentMenuPtr(core_ptr);
	if(m == NULL)
	    return;

        /* Handle by spin's ID code. Also note that the value's
         * index position is used, not the actual value so the
         * order of values placed into the spin object is important.
         */
        switch(id_code)
        {
          /* ****************************************************** */
          case SAR_OPT_SELECT_HOIST_CONTACT:
            /* These values should correspond to those used to
             * set the values in SARBuildMenus().
             */
            switch(spin_ptr->cur_value)
            {
	      case 1:	/* Moderate. */
		option.hoist_contact_expansion_coeff = 2.0;
		break;

	      case 2:	/* Authentic/difficult. */
		option.hoist_contact_expansion_coeff = 1.0;
		break;

	      default:	/* Easy. */
		option.hoist_contact_expansion_coeff = 4.0;
		break;
	    }
	    break;

          /* ****************************************************** */
          case SAR_OPT_SELECT_DAMAGE_RESISTANCE:
            /* These values should correspond to those used to
             * set the values in SARBuildMenus().
             */
            switch(spin_ptr->cur_value)
            {
              case 1:   /* Nominal. */
		option.damage_resistance_coeff = 1.5;
                break;

              case 2:   /* Authentic/difficult. */
		option.damage_resistance_coeff = 1.0;
                break;
                    
              default:	/* Strong/easy. */
		option.damage_resistance_coeff = 2.0;
                break;
            }   
            break;

          /* ****************************************************** */
          case SAR_OPT_SELECT_FLIGHT_PHYSICS:
            /* These values should correspond to those used to
             * set the values in SARBuildMenus().
             */
            switch(spin_ptr->cur_value)
            {
              case 1:   /* Moderate. */
		option.flight_physics = FLIGHT_PHYSICS_MODERATE;
                break;

              case 2:   /* Realistic/difficult. */
		option.flight_physics = FLIGHT_PHYSICS_REALISTIC;
                break;

              default:  /* Easy. */
		option.flight_physics = FLIGHT_PHYSICS_EASY;
                break;
            }
            break;

          /* ****************************************************** */
          case SAR_OPT_SELECT_JS0_CONNECTION:
            /* These values should correspond to those used to
             * set the values in SARBuildMenus().
             */
#define CONNECTION_TYPE	option.js0_connection
            switch(spin_ptr->cur_value)
            {
              case 1:   /* USB. */
                CONNECTION_TYPE = GCTL_JS_CONNECTION_USB;
		break;
	      default:	/* Standard. */
		CONNECTION_TYPE = GCTL_JS_CONNECTION_STANDARD;
                break;
	    }
#undef CONNECTION_TYPE
	    /* Update options.controllers mask and reinitialize the
	     * game controller.
	     */
	    SARMenuOptionsReinitJoysticks(core_ptr);
	    break;

          /* ****************************************************** */
          case SAR_OPT_SELECT_JS1_CONNECTION:
            /* These values should correspond to those used to
             * set the values in SARBuildMenus().
             */
#define CONNECTION_TYPE option.js1_connection
            switch(spin_ptr->cur_value)
            {
              case 1:   /* USB. */
                CONNECTION_TYPE = GCTL_JS_CONNECTION_USB;
                break;
              default:  /* Standard. */
                CONNECTION_TYPE = GCTL_JS_CONNECTION_STANDARD;
                break;
            }
#undef CONNECTION_TYPE
            /* Update options.controllers mask and reinitialize the
             * game controller.
             */
            SARMenuOptionsReinitJoysticks(core_ptr);
            break;

          /* ****************************************************** */
          case SAR_OPT_SELECT_JS0_AXISES:
            /* These values should correspond to those used to
             * set the values in SARBuildMenus().
             */
#define AXIS_ROLES	option.gctl_js0_axis_roles
	    AXIS_ROLES = 0;
            switch(spin_ptr->cur_value)
            {
              case 1:   /* 2D. */
                AXIS_ROLES |= GCTL_JS_AXIS_ROLE_PITCH |
		    GCTL_JS_AXIS_ROLE_BANK;
                break;

              case 2:   /* 2D with throttle. */ 
                AXIS_ROLES |= GCTL_JS_AXIS_ROLE_PITCH |
		    GCTL_JS_AXIS_ROLE_BANK | GCTL_JS_AXIS_ROLE_THROTTLE;
                break;

              case 3:   /* 2D with hat. */
                AXIS_ROLES |= GCTL_JS_AXIS_ROLE_PITCH |
		    GCTL_JS_AXIS_ROLE_BANK | GCTL_JS_AXIS_ROLE_HAT;
                break;

              case 4:   /* 2D with throttle & hat. */
                AXIS_ROLES |= GCTL_JS_AXIS_ROLE_PITCH |
		    GCTL_JS_AXIS_ROLE_BANK | GCTL_JS_AXIS_ROLE_THROTTLE |
		    GCTL_JS_AXIS_ROLE_HAT;
                break;
            
              case 5:   /* 3D. */
                AXIS_ROLES |= GCTL_JS_AXIS_ROLE_PITCH |
		    GCTL_JS_AXIS_ROLE_BANK | GCTL_JS_AXIS_ROLE_HEADING;
                break;
                
              case 6:   /* 3D with throttle. */
                AXIS_ROLES |= GCTL_JS_AXIS_ROLE_PITCH |
		    GCTL_JS_AXIS_ROLE_BANK | GCTL_JS_AXIS_ROLE_HEADING |
		    GCTL_JS_AXIS_ROLE_THROTTLE;
                break;
        
              case 7:   /* 3D with hat. */
                AXIS_ROLES |= GCTL_JS_AXIS_ROLE_PITCH |
		    GCTL_JS_AXIS_ROLE_BANK | GCTL_JS_AXIS_ROLE_HEADING |
		    GCTL_JS_AXIS_ROLE_HAT;
                break;
              
              case 8:   /* 3D with throttle & hat. */
                AXIS_ROLES |= GCTL_JS_AXIS_ROLE_PITCH |
		    GCTL_JS_AXIS_ROLE_BANK | GCTL_JS_AXIS_ROLE_HEADING |
		    GCTL_JS_AXIS_ROLE_THROTTLE | GCTL_JS_AXIS_ROLE_HAT;
                break;
                
              case 9:   /* As throttle & rudder. */
                AXIS_ROLES |= GCTL_JS_AXIS_ROLE_AS_THROTTLE_AND_RUDDER;
                break;
            }
#undef AXIS_ROLES
            /* Update options.controllers mask and reinitialize the
             * game controller.
             */
            SARMenuOptionsReinitJoysticks(core_ptr);
            break;
                    
          /* ****************************************************** */
          case SAR_OPT_SELECT_JS1_AXISES:
            /* These values should correspond to those used to
             * set the values in SARBuildMenus().
             */
#define AXIS_ROLES	option.gctl_js1_axis_roles
            AXIS_ROLES = 0;
            switch(spin_ptr->cur_value)
            {
              case 1:   /* 2D. */
                AXIS_ROLES |= GCTL_JS_AXIS_ROLE_PITCH |
                    GCTL_JS_AXIS_ROLE_BANK;
                break;

              case 2:   /* 2D with throttle. */
                AXIS_ROLES |= GCTL_JS_AXIS_ROLE_PITCH |
                    GCTL_JS_AXIS_ROLE_BANK | GCTL_JS_AXIS_ROLE_THROTTLE;
                break;

              case 3:   /* 2D with hat. */
                AXIS_ROLES |= GCTL_JS_AXIS_ROLE_PITCH |
                    GCTL_JS_AXIS_ROLE_BANK | GCTL_JS_AXIS_ROLE_HAT;
                break;

              case 4:   /* 2D with throttle & hat. */
                AXIS_ROLES |= GCTL_JS_AXIS_ROLE_PITCH |
                    GCTL_JS_AXIS_ROLE_BANK | GCTL_JS_AXIS_ROLE_THROTTLE |
                    GCTL_JS_AXIS_ROLE_HAT;
                break;

              case 5:   /* 3D. */
                AXIS_ROLES |= GCTL_JS_AXIS_ROLE_PITCH |
                    GCTL_JS_AXIS_ROLE_BANK | GCTL_JS_AXIS_ROLE_HEADING;
                break;

              case 6:   /* 3D with throttle. */
                AXIS_ROLES |= GCTL_JS_AXIS_ROLE_PITCH |
                    GCTL_JS_AXIS_ROLE_BANK | GCTL_JS_AXIS_ROLE_HEADING |
                    GCTL_JS_AXIS_ROLE_THROTTLE;
                break;

              case 7:   /* 3D with hat. */
                AXIS_ROLES |= GCTL_JS_AXIS_ROLE_PITCH |
                    GCTL_JS_AXIS_ROLE_BANK | GCTL_JS_AXIS_ROLE_HEADING |
                    GCTL_JS_AXIS_ROLE_HAT;
                break;

              case 8:   /* 3D with throttle & hat. */
                AXIS_ROLES |= GCTL_JS_AXIS_ROLE_PITCH |
                    GCTL_JS_AXIS_ROLE_BANK | GCTL_JS_AXIS_ROLE_HEADING |
                    GCTL_JS_AXIS_ROLE_THROTTLE | GCTL_JS_AXIS_ROLE_HAT;
                break;

              case 9:   /* As throttle & rudder. */
                AXIS_ROLES |= GCTL_JS_AXIS_ROLE_AS_THROTTLE_AND_RUDDER;
                break;
            }
#undef AXIS_ROLES
            /* Update options.controllers mask and reinitialize the
             * game controller.
             */
            SARMenuOptionsReinitJoysticks(core_ptr);
            break;

          /* ****************************************************** */
          case SAR_OPT_SELECT_JS0_BUTTON_ACTION:
#ifdef JS_SUPPORT
	    if(1)
	    {
		int n;
		sar_menu_spin_struct *btn_spin_ptr;

		btn_spin_ptr = SAROptionsGetSpinByID(
		    m, SAR_OPT_SELECT_JS0_BUTTON_NUMBER, &n
		);
		if(btn_spin_ptr != NULL)
		{
		    btn_spin_ptr->cur_value = SARGetJSButtonFromSpinAction(
			spin_ptr, SAR_OPT_SELECT_JS0_BUTTON_ACTION
		    ) + 1;
                    if(btn_spin_ptr->cur_value >= btn_spin_ptr->total_values)
                        btn_spin_ptr->cur_value = 0;
		    SARMenuDrawObject(core_ptr->display, m, n);
		    GWSwapBuffer(core_ptr->display);
		}
	    }
#endif	/* JS_SUPPORT. */
	    break;

          /* ****************************************************** */
          case SAR_OPT_SELECT_JS0_BUTTON_NUMBER:
#ifdef JS_SUPPORT
            if(1)
            {
		sar_option_struct *opt = &option;
                sar_menu_spin_struct *act_spin_ptr = SAROptionsGetSpinByID(
                    m, SAR_OPT_SELECT_JS0_BUTTON_ACTION, NULL
                );
		if(act_spin_ptr != NULL)
		{
		    int new_val = spin_ptr->cur_value - 1;

		    /* Set option for button mapping based on which value
		     * the action spin is on.
		     */
		    switch(act_spin_ptr->cur_value)
		    {
		      case 0:	/* Rotate Modifier. */
			opt->js0_btn_rotate = new_val;
			break;
                      case 1:	/* Air Brakes. */
                        opt->js0_btn_air_brakes = new_val;
                        break;
                      case 2:	/* Wheel Brakes. */
                        opt->js0_btn_wheel_brakes = new_val;
                        break;
                      case 3:	/* Zoom In. */
                        opt->js0_btn_zoom_in = new_val;
                        break;
                      case 4:	/* Zoom out. */
                        opt->js0_btn_zoom_out = new_val;
                        break;
                      case 5:	/* Hoist Up. */
                        opt->js0_btn_hoist_up = new_val;
                        break;
                      case 6:	/* Hoist Down. */
                        opt->js0_btn_hoist_down = new_val;
                        break;
/* When adding new button mappings, be sure to add support for them
 * in SARGetJSButtonFromSpinAction().
 *
 * Also remember to add support for EACH joystick!
 */
		    }
                }
            }
#endif  /* JS_SUPPORT. */
	    /* Update joystick button mappings from global options
	     * values to the gctl structure on the core structure.
	     * No reinitialization will take place.
	     */
	    SARMenuOptionsRemapJoystickButtons(core_ptr);
            break;

          /* ****************************************************** */
          case SAR_OPT_SELECT_JS1_BUTTON_ACTION:
#ifdef JS_SUPPORT
            if(1)
            {
                int n;
                sar_menu_spin_struct *btn_spin_ptr;

                btn_spin_ptr = SAROptionsGetSpinByID(
                    m, SAR_OPT_SELECT_JS1_BUTTON_NUMBER, &n 
                );
                if(btn_spin_ptr != NULL)
                {
                    btn_spin_ptr->cur_value = SARGetJSButtonFromSpinAction(
                        spin_ptr, SAR_OPT_SELECT_JS1_BUTTON_ACTION
                    ) + 1;
                    if(btn_spin_ptr->cur_value >= btn_spin_ptr->total_values)
                        btn_spin_ptr->cur_value = 0;
                    SARMenuDrawObject(core_ptr->display, m, n);
                    GWSwapBuffer(core_ptr->display);
                }
            }
#endif  /* JS_SUPPORT. */
            break;

          /* ****************************************************** */
          case SAR_OPT_SELECT_JS1_BUTTON_NUMBER:
#ifdef JS_SUPPORT
            if(1)
            {
                sar_option_struct *opt = &option;
                sar_menu_spin_struct *act_spin_ptr = SAROptionsGetSpinByID(
                    m, SAR_OPT_SELECT_JS1_BUTTON_ACTION, NULL
                );
                if(act_spin_ptr != NULL)
                {
                    int new_val = spin_ptr->cur_value - 1;

                    /* Set option for button mapping based on which value
                     * the action spin is on.
                     */
                    switch(act_spin_ptr->cur_value)
                    {
                      case 0:   /* Rotate Modifier. */
                        opt->js1_btn_rotate = new_val;
                        break;
                      case 1:   /* Air Brakes. */
                        opt->js1_btn_air_brakes = new_val;
                        break;
                      case 2:   /* Wheel Brakes. */
                        opt->js1_btn_wheel_brakes = new_val;
                        break;
                      case 3:   /* Zoom In. */
                        opt->js1_btn_zoom_in = new_val;
                        break;
                      case 4:   /* Zoom out. */
                        opt->js1_btn_zoom_out = new_val;
                        break;
                      case 5:   /* Hoist Up. */
                        opt->js1_btn_hoist_up = new_val;
                        break;
                      case 6:   /* Hoist Down. */
                        opt->js1_btn_hoist_down = new_val;
                        break;
/* When adding new button mappings, be sure to add support for them
 * in SARGetJSButtonFromSpinAction().
 * 
 * Also remember to add support for EACH joystick!
 */
                    }
                }
            }
#endif  /* JS_SUPPORT. */
            /* Update joystick button mappings from global options
             * values to the gctl structure on the core structure.
             * No reinitialization will take place.
             */
            SARMenuOptionsRemapJoystickButtons(core_ptr);
            break;

          /* ****************************************************** */
	  case SAR_OPT_SELECT_MAX_VISIBILITY:
            /* These values should correspond to those used to
             * set the values in SARBuildMenus().
             */
	    option.visibility_max = spin_ptr->cur_value;
	    break;

          /* ****************************************************** */
          case SAR_OPT_SELECT_SOUND_LEVEL:
/* Macro to turn sound on as needed by checking if the core structure's
 * recorder is initialized. If the recorder is NULL then an attempt to
 * connect to the sound server will be made using the argument
 * specified by sound_server_connect_arg.
 */
#define SND_ON_AS_NEEDED        \
if(core_ptr->recorder == NULL) \
{ \
 GWSetInputBusy(core_ptr->display); \
 core_ptr->recorder = SoundInit( \
  SNDSERV_TYPE_Y, \
  sound_server_connect_arg, \
  NULL		/* Do not start sound server. */ \
 ); \
 if(core_ptr->recorder == NULL) \
 { \
  char *buf; \
  int buf_len; \
 \
  buf_len = 512 + ((sound_server_connect_arg == NULL) ? \
   0 : strlen(sound_server_connect_arg) \
  ); \
  buf = (char *)malloc(buf_len * sizeof(char)); \
  if(buf != NULL) \
   sprintf(buf, \
"Unable to connect to sound server at:\n\n    %s\n\n\
Make sure that the address is correct,\n\
use the argument --recorder to specify\n\
an alternate address as needed.\n\n\
To start the sound server in most cases\n\
use `starty'", \
    sound_server_connect_arg \
   ); \
  GWOutputMessage( \
   core_ptr->display, \
   GWOutputMessageTypeError, \
   "Sound initialization failed!", \
   buf, \
"Please check to make sure that your sound server\n\
is running and available at the specified address.\n\
Also make sure that you have sufficient permission\n\
to connect to it. To start in most cases run `starty'." \
   ); \
   free(buf); \
 } \
 else \
 { \
  SoundChangeMode( \
   core_ptr->recorder, core_ptr->audio_mode_name \
  ); \
 } \
 GWSetInputReady(core_ptr->display); \
}

#define SND_OFF_AS_NEEDED       \
if(core_ptr->recorder != NULL) \
{ \
 GWSetInputBusy(core_ptr->display); \
 SoundShutdown(core_ptr->recorder); \
 core_ptr->recorder = NULL; \
 GWSetInputReady(core_ptr->display); \
}

            switch(spin_ptr->cur_value)
            {
              case 1:   /* On, events only. */
                option.engine_sounds = False;
		option.event_sounds = True;
		option.voice_sounds = False;
                SND_ON_AS_NEEDED
                break;

              case 2:   /* On, events and engine. */
                option.engine_sounds = True;
		option.event_sounds = True;
		option.voice_sounds = False;
                SND_ON_AS_NEEDED  
                break;
 
              case 3:   /* On, events, engine, and voice (all on). */
		option.engine_sounds = True;
		option.event_sounds = True;
		option.voice_sounds = True;
		SND_ON_AS_NEEDED
                break;
    
              default:  /* All else assume off. */
		option.engine_sounds = False;
		option.event_sounds = False;
		option.voice_sounds = False;
                SND_OFF_AS_NEEDED
                break;
            }
#undef SND_ON_AS_NEEDED

#if !defined(HAVE_Y2)
	    /* Print warning about no sound support. */
	    GWOutputMessage(
		core_ptr->display, GWOutputMessageTypeWarning,
		"Sound Support Not Available",
"There is no sound support available.",
"This program was not compiled with sound support and/or\n\
sound support was not enabled on your operating system."
	    );
#endif

            break;

        }
}
